/*
 * HP48G GROB/GG/4C/GA0 File Viewer
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#ifdef linux
#include <vga.h>
#endif

#ifdef linux
int vga_kbhit(void)
{
    return 0;
}
#endif

#ifdef __WATCOMC__

#include <conio.h>
#include <i86.h>

#define G320x200x256 0x13
#define TEXT         0x03

int vga_color;

void vga_setmode(int mode)
{
    union REGS r;
    r.x.eax = mode;
    int386(0x10, &r, &r);
}

void vga_setpalette(int index, int red, int green, int blue)
{
    outp(0x3c8, index);
    outp(0x3c9, red);
    outp(0x3c9, green);
    outp(0x3c9, blue);
}

void vga_setcolor(int color)
{
    vga_color = color;
}

void vga_drawpixel(int x, int y)
{
    *((char*)(0xa0000 + x + 320 * y)) = vga_color;
}

int vga_getpixel(int x, int y)
{
    return *((char*)(0xa0000 + x + 320 * y));
}

void vga_getch(void)
{
    getch();
}

int vga_kbhit(void)
{
    return kbhit();
}

#endif

/* HP48 file magic cookies */
#define HP48MAGIC   "HPHP48-"
#define HP48PROLOG  0x02B1EL

/* supported HP48 file formats */
#define HP48GROB    0
#define HP48GG      1
#define HP484C      2
#define HP48GA0     3

int main(int argc, char *argv[])
{
    FILE *fp;
    unsigned char hdr[18];
    char *ext, *ptr, *pixmap;
    long prolog, length, width, height, slides, frames;
    int x, y, z, t, c, color, type;
    clock_t time0;

    /* print banner */
    printf("HP48G GROB/GG/4C/GA0 Viewer Version 0.01 08-02-96\n");
    printf("Copyright (C) 1996 Pel Software. Freeware Version\n");

    /* check command line arguments */
    if (argc != 2) {
        fprintf(stderr, "Usage: GV file.[grb|gro|gg|4c|ga0]\n");
        exit(1);
    }

    /* open HP48 file stream */
    if ((fp = fopen(argv[1], "rb")) == NULL) {
        fprintf(stderr, "ERROR: File Not Found\n");
        exit(1);
    }

    /* determine HP48 file format by extension */
    if ((ext = strchr(argv[1], '.')) == NULL)
        type = HP48GROB;
    else if (!strcmp(ext, ".GRB") || !strcmp(ext, ".grb"))
        type = HP48GROB;
    else if (!strcmp(ext, ".GRO") || !strcmp(ext, ".gro"))
        type = HP48GROB;
    else if (!strcmp(ext, ".GG")  || !strcmp(ext, ".gg"))
        type = HP48GG;
    else if (!strcmp(ext, ".4C")  || !strcmp(ext, ".4c"))
        type = HP484C;
    else if (!strcmp(ext, ".GA0") || !strcmp(ext, ".ga0"))
        type = HP48GA0;
    else {
        fprintf(stderr, "ERROR: Unknown File Extension\n");
        exit(1);
    }

    /* read HP48 header structure */
    if (fread(hdr, sizeof(char), sizeof(hdr), fp) != sizeof(hdr)) {
        fprintf(stderr, "ERROR: Can't Read HP48 Header Structure\n");
        exit(1);
    }

    /* verify HP48 header signature */
    if (memcmp(hdr, HP48MAGIC, 7)) {
        fprintf(stderr, "ERROR: Not an HP48 Binary File\n");
        exit(1);
    }

    /* grab and unpack HP48 header field values */
    prolog = hdr[8] | ((long) hdr[9] << 8) | ((long) (hdr[10] & 0x0f) << 16);
    length = (hdr[10] >> 4) | ((long) hdr[11] << 4) | ((long) hdr[12] << 12);
    height = hdr[13] | ((long) hdr[14] << 8) | ((long) (hdr[15] & 0x0f) << 16);
    width = (hdr[15] >> 4) | ((long) hdr[16] << 4) | ((long) hdr[17] << 12);

    /* check HP48 binary file prolog signature */
    if (prolog != HP48PROLOG) {
        fprintf(stderr, "ERROR: Not an HP48 GROB Binary File\n");
        exit(1);
    }

    /* safety check HP48/GROB length field */
    if (length != 15 + height * (2 * ((width + 7) / 8))) {
        fprintf(stderr, "ERROR: Corrupted HP48 GROB Header Fields\n");
        exit(1);
    }

    /* determine number of slides */
    if (type == HP48GROB) {
        slides = 1;
        frames = 1;
    }
    else if (type == HP48GG) {
        if (width != 131 || (height % 64))
            fprintf(stderr, "WARNING: Non Standard HP48 GrayGrob File\n");
        slides = height / 64;
        height = 64;
        frames = 1;
    }
    else if (type == HP484C) {
        if (height % 2)
            fprintf(stderr, "WARNING: Non Standard HP48 FourColor File\n");
        slides = 2;
        height = height / 2;
        frames = 1;
    }
    else if (type == HP48GA0) {
        if (width != 128 || (height % 112))
            fprintf(stderr, "WARNING: Non Standard HP48 GrobAnim File\n");
        slides = 2;
        frames = height / 112;
        height = 56;
    }

    /* show HP48 file information */
    printf("Loading %ldx%ld HP48 %s\n", width, height,
        type == HP48GROB ? "GROB Image" :
        type == HP48GG   ? "GrayGrob Image" :
        type == HP484C   ? "FourColor Image" :
        type == HP48GA0  ? "GROB Animation" : NULL);

    /* allocate space for pixmaps */
    if ((pixmap = calloc(frames, width * height)) == NULL) {
        fprintf(stderr, "ERROR: Not Enough Memory\n");
        exit(1);
    }

    /* read HP48 GROB bitmap slides */
    for (t = 0; t < frames; t++) {
        for (z = 0; z < slides; z++) {
            color = 1;
            if (type == HP484C || type == HP48GA0)
                color = 1 << (slides - z - 1);
            for (y = 0; y < height; y++) {
                ptr = &pixmap[width * (y + height * t)];
                for (x = 0; x < width; x += 8) {
                    if ((c = fgetc(fp)) == EOF) {
                        fprintf(stderr, "ERROR: Truncated HP48 Binary File\n");
                        exit(1);
                    }
                    if (!(c & 0x01) && x + 0 < width) ptr[0] += color;
                    if (!(c & 0x02) && x + 1 < width) ptr[1] += color;
                    if (!(c & 0x04) && x + 2 < width) ptr[2] += color;
                    if (!(c & 0x08) && x + 3 < width) ptr[3] += color;
                    if (!(c & 0x10) && x + 4 < width) ptr[4] += color;
                    if (!(c & 0x20) && x + 5 < width) ptr[5] += color;
                    if (!(c & 0x40) && x + 6 < width) ptr[6] += color;
                    if (!(c & 0x80) && x + 7 < width) ptr[7] += color;
                    ptr += 8;
                }
            }
        }
    }

    fclose(fp);

    /* show pixmap on screen */
    vga_setmode(G320x200x256);
    if (type == HP484C || type == HP48GA0)
        slides = (1 << slides) - 1;
    for (z = 0; z <= slides; z++) {
        c = (63 * z) / slides;
        vga_setpalette(z, c, c, c);
    }
    while (!vga_kbhit()) {
        ptr = pixmap;
        time0 = clock();
        for (t = 0; t < frames; t++) {
            for (y = 0; y < height; y++) {
                for (x = 0; x < width; x++) {
                    vga_setcolor(*ptr++);
                    vga_drawpixel(x, y);
                }
            }
            time0 += (100L * CLOCKS_PER_SEC) / 1000;
            while (time0 > clock())
                continue;
            if (vga_kbhit()) break;
        }
    }
    vga_getch();
    vga_setmode(TEXT);

    /* release pixmap and exit */
    free(pixmap);
    return 0;
}
