/*
 * HP48G GROB/GG/4C/GA0 File Viewer
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
/* if you are on a sun/solaris machine you probably need too specify
* where to find X11 by compiling with -I/store/include -L/store/lib or 
* some similar */

/* HP48 file magic cookies */
#define HP48MAGIC   "HPHP48-"
#define HP48PROLOG  0x02B1EL

/* supported HP48 file formats */
#define HP48GROB    0
#define HP48GG      1
#define HP484C      2
#define HP48GA0     3

Display *dpy;
Window win, dummy;
int screen, depth;
GC *gc;						  /* one for each slide */
Pixmap *pix=0;					  /* one for eace frame */
XColor *xcolor;					  /* one for each color */
int win_width, win_height;
unsigned int d, d_w, d_h;
long width, height, frames;
char *pixmap;

void display(int t, XExposeEvent *event);
void make_pix(int t, XConfigureEvent *event);
void XDelay(Display *display,unsigned long milliseconds);

int main(int argc, char *argv[])
{
  FILE *fp;
  unsigned char hdr[18];
  char *ext, *ptr;
  long prolog, length, slides;
  int x, y, z, t, color, c, type, done=0;
  clock_t time0;
  
  /* X-stuff */
  char  *progname, keybuf[10];
  XSizeHints hint;
  XEvent xev;
  KeySym key;
  Visual *visual;
  Colormap cmap;
  int r, g, b, contrast;
  int frame;
  
  /* Get the name we are called. */
  progname = strrchr(argv[0], '/');
  if (progname == NULL)
    progname = argv[0];
  else
    progname++;
  
  /* print banner */
  printf("HP48G GROB/GG/4C/GA0 Viewer Version 0.01 08-02-96\n");
  printf("Copyright (C) 1996 Pel Software. Freeware Version\n");

  /* check command line arguments */
  if (argc != 2) {
    fprintf(stderr, "Usage: %s file.[grb|gro|gg|4c|ga0]\n", progname);
    exit(1);
  }

  /* open HP48 file stream */
  if ((fp = fopen(argv[1], "rb")) == NULL) {
    fprintf(stderr, "ERROR: File Not Found\n");
    exit(1);
  }
  
  /* determine HP48 file format by extension */
  if ((ext = strchr(argv[1], '.')) == NULL)
    type = HP48GROB;
  else if (!strcmp(ext, ".GRB") || !strcmp(ext, ".grb"))
    type = HP48GROB;
  else if (!strcmp(ext, ".GRO") || !strcmp(ext, ".gro"))
    type = HP48GROB;
  else if (!strcmp(ext, ".GG")  || !strcmp(ext, ".gg"))
    type = HP48GG;
  else if (!strcmp(ext, ".4C")  || !strcmp(ext, ".4c"))
    type = HP484C;
  else if (!strcmp(ext, ".GA0") || !strcmp(ext, ".ga0"))
    type = HP48GA0;
  else {
    fprintf(stderr, "ERROR: Unknown File Extension\n");
    exit(1);
  }

  /* read HP48 header structure */
  if (fread(hdr, sizeof(char), sizeof(hdr), fp) != sizeof(hdr)) {
    fprintf(stderr, "ERROR: Can't Read HP48 Header Structure\n");
    exit(1);
  }

  /* verify HP48 header signature */
  if (memcmp(hdr, HP48MAGIC, 7)) {
    fprintf(stderr, "ERROR: Not an HP48 Binary File\n");
    exit(1);
  }

  /* grab and unpack HP48 header field values */
  prolog = hdr[8] | ((long) hdr[9] << 8) | ((long) (hdr[10] & 0x0f) << 16);
  length = (hdr[10] >> 4) | ((long) hdr[11] << 4) | ((long) hdr[12] << 12);
  height = hdr[13] | ((long) hdr[14] << 8) | ((long) (hdr[15] & 0x0f) << 16);
  width = (hdr[15] >> 4) | ((long) hdr[16] << 4) | ((long) hdr[17] << 12);

    /* check HP48 binary file prolog signature */
  if (prolog != HP48PROLOG) {
    fprintf(stderr, "ERROR: Not an HP48 GROB Binary File\n");
    exit(1);
  }

  /* safety check HP48/GROB length field */
  if (length != 15 + height * (2 * ((width + 7) / 8))) {
    fprintf(stderr, "ERROR: Corrupted HP48 GROB Header Fields\n");
    exit(1);
  }

  /* determine number of slides */
  if (type == HP48GROB) {
    slides = 1;
    frames = 1;
  }
  else if (type == HP48GG) {
    if (width != 131 || (height % 64))
      fprintf(stderr, "WARNING: Non Standard HP48 GrayGrob File\n");
    slides = height / 64;
    height = 64;
    frames = 1;
  }
  else if (type == HP484C) {
    if (height % 2)
      fprintf(stderr, "WARNING: Non Standard HP48 FourColor File\n");
    slides = 2;
    height = height / 2;
    frames = 1;
  }
  else if (type == HP48GA0) {
    if (width != 128 || (height % 112))
      fprintf(stderr, "WARNING: Non Standard HP48 GrobAnim File\n");
    slides = 2;
    frames = height / 112;
    height = 56;
  }

  /* show HP48 file information */
  printf("Loading %ldx%ld HP48 %s\n", width, height,
	 type == HP48GROB ? "GROB Image" :
	 type == HP48GG   ? "GrayGrob Image" :
	 type == HP484C   ? "FourColor Image" :
	 type == HP48GA0  ? "GROB Animation" : NULL);

  /* allocate space for pixmaps */
  if ((pixmap = calloc(frames, width * height)) == NULL) {
    fprintf(stderr, "ERROR: Not Enough Memory\n");
    exit(1);
  }

  /* read HP48 GROB bitmap slides */
  for (t = 0; t < frames; t++) {
    for (z = 0; z < slides; z++) {
      color = 1;
      if (type == HP484C || type == HP48GA0)
	color = 1 << (slides - z - 1);
      for (y = 0; y < height; y++) {
	ptr = &pixmap[width * (y + height * t)];
	for (x = 0; x < width; x += 8) {
	  if ((c = fgetc(fp)) == EOF) {
	    fprintf(stderr, "ERROR: Truncated HP48 Binary File\n");
	    exit(1);
	  }
	  if (!(c & 0x01) && x + 0 < width) ptr[0] += color;
	  if (!(c & 0x02) && x + 1 < width) ptr[1] += color;
	  if (!(c & 0x04) && x + 2 < width) ptr[2] += color;
	  if (!(c & 0x08) && x + 3 < width) ptr[3] += color;
	  if (!(c & 0x10) && x + 4 < width) ptr[4] += color;
	  if (!(c & 0x20) && x + 5 < width) ptr[5] += color;
	  if (!(c & 0x40) && x + 6 < width) ptr[6] += color;
	  if (!(c & 0x80) && x + 7 < width) ptr[7] += color;
	  ptr += 8;
	}
      }
    }
  }

  fclose(fp);
  if (type == HP484C || type == HP48GA0)
    slides = (1 << slides) - 1;

  /*
   * X - stuff
   */
  /* connect to the server */
  if ( (dpy = XOpenDisplay("")) == NULL ){
    fprintf(stderr,"%s: cannot connect to X server %s\n",
	    progname, XDisplayName(""));
    exit(EXIT_FAILURE);
  }
  /* get the default */
  screen = DefaultScreen(dpy);
  depth = DefaultDepth(dpy, screen);
  visual = DefaultVisual(dpy, screen);

  /* create the GC's */
  if ( (gc = calloc(slides + 1, sizeof(gc))) == NULL ) {
    fprintf(stderr, "ERROR: Not Enough Memory\n");
    exit(EXIT_FAILURE);
  }
/*   gc = DefaultGC(dpy, screen); */
  
  /* allocate colors */
  if ( (xcolor = calloc(slides + 1, sizeof(XColor))) == NULL ) {
    fprintf(stderr, "ERROR: Not Enough Memory\n");
    exit(EXIT_FAILURE);
  }
  cmap = DefaultColormap(dpy, screen);    
  xcolor[0].red = 0 << 8;
  xcolor[0].green = 0 << 8;
  xcolor[0].blue = 128 << 8;
  xcolor[slides].red = 202 << 8;
  xcolor[slides].green = 221 << 8;
  xcolor[slides].blue = 92 << 8;
  for (z = 1; z < slides; z++) {
    contrast = z*(0x13/slides);
    r = (contrast) * (202 / 0x10);
    g = (contrast) * (221 / 0x10);
    b = 128 - ((contrast) * ((128 - 92) / 0x10));
    xcolor[z].red = r << 8;
    xcolor[z].green = g << 8;
    xcolor[z].blue = b << 8;
  }
  
  for (z = 0; z <= slides; z++){
    if (XAllocColor(dpy, cmap, &xcolor[z]) == NULL)
      fprintf(stderr, "%s: XAllocColor failed, color %d", progname, z);
    gc[z] = XCreateGC(dpy, DefaultRootWindow(dpy), 0, (XGCValues*) NULL);
    XSetForeground(dpy, gc[z], xcolor[z].pixel);    
    XSetBackground(dpy, gc[z], xcolor[slides].pixel);
    }

  /* create the display window */
  win_width = width;
  win_height = height;
  if((win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
				0, 0, win_width, win_height, 0,
				xcolor[0].pixel, xcolor[slides].pixel))
     == NULL ) {
    fprintf(stderr, "cannot open window\n");
    exit(EXIT_FAILURE);
  }

  /* allocate X-pixmaps */
  if ( (pix = calloc(frames, sizeof(Pixmap))) == NULL) {
    fprintf(stderr, "ERROR: Not Enough Memory\n");
    exit(EXIT_FAILURE);
  }
  xev.xconfigure.width = win_width;
  xev.xconfigure.height = win_height;
  for (t = 0; t < frames; t++) 
    make_pix(t, (XConfigureEvent *) &xev);
    
  /* set property hints */
  hint.flags = PPosition | PSize | PMinSize | PResizeInc | PAspect;
  hint.x = hint.y = 0;
  hint.width = hint.min_width = width;
  hint.height =hint.min_height = height;
  hint.width_inc = width;
  hint.height_inc = height;
  hint.min_aspect.x = hint.max_aspect.x = width;
  hint.min_aspect.y = hint.max_aspect.y = height;
  XSetStandardProperties(dpy, win, progname, progname, None,
			 argv, argc, &hint);
  /*   XSetWMNormalHints(dpy, win, &hint); */

  /* select input and raise window */
  XSelectInput(dpy, win, KeyPressMask | ExposureMask | StructureNotifyMask);
  XMapRaised(dpy, win);

  /* show pixmap on screen */
  t = 0;
  while(!done) {
    while(XPending(dpy)) {
      XNextEvent(dpy, &xev);
      switch(xev.type) {
      case KeyPress:
	c = XLookupString(&xev.xkey, keybuf, 10, &key, 0);
	if(c==1 && *keybuf=='q')
	  done = 1;
	break;
      case Expose:
	display(t,(XExposeEvent *) &xev);
	break;
      case ConfigureNotify:
	if ((xev.xconfigure.width != win_width) ||
	    (xev.xconfigure.height != win_height)) {
	  win_width = xev.xconfigure.width;
	  win_height = xev.xconfigure.height;
	}
	break;
      }
    }
    if (++t >= frames)
      t = 0;
    XGetGeometry(dpy, pix[t], &dummy, (int*)&d, (int*)&d, &d_w, &d_h, &d, &d);
    if ((d_w != win_width) || (d_h != win_height))
      make_pix(t, (XConfigureEvent *) &xev);
    else
      XDelay(dpy, 150L);
    xev.xexpose.x = 0;
    xev.xexpose.y = 0;
    xev.xexpose.width = win_width;
    xev.xexpose.height = win_height;
    display(t, (XExposeEvent *) &xev);
  }
  XFlush(dpy);
  XDestroyWindow(dpy, win);
  XCloseDisplay(dpy);
/* #endif */
    
  /* release pixmap and exit */

   free(pixmap);
}

void display(int t, XExposeEvent *event)
{
  /*
   * Extract the exposed area from the event and copy
   * from the saved pixmap to the window.
   */
  XCopyArea(dpy, pix[t], win, gc[0], 
	    event->x, event->y, event->width, event->height, 
	    event->x, event->y);
}

void make_pix(int t, XConfigureEvent *event)
{
  int x, y, zoom;
  char *ptr;
  
  if (pix[t])
    XFreePixmap(dpy, pix[t]);
  
  zoom = win_width/width;
  ptr = pixmap+(t*width*height);
  pix[t] = XCreatePixmap(dpy, win, win_width, win_height, depth);
  for (y = 0; y < height; y++)
    for (x = 0; x < width; x++) 
      XFillRectangle(dpy, pix[t], gc[*ptr++], x*zoom, y*zoom, zoom, zoom);
}

void XDelay(Display *display,unsigned long milliseconds)
{
  XFlush(display);
  if (milliseconds == 0)
    return;
#ifdef sysv
  {
#include <stropts.h>
#include <poll.h>
    
    (void) poll((struct pollfd *) NULL,0,(int) milliseconds);
  }
#else
  {
#include <unistd.h>
#include <sys/types.h>
#include <bstring.h>
#include <sys/time.h>
    
    struct timeval timer;

    timer.tv_sec=milliseconds/1000;
    timer.tv_usec=(milliseconds % 1000)*1000;
    (void) select(0,(fd_set *) NULL,(fd_set *) NULL,(fd_set *) NULL,&timer);
  }
#endif
}
